package net.sergeych.kiloparsec.adapter

import io.ktor.network.selector.*
import io.ktor.network.sockets.*
import io.ktor.util.network.*
import kotlinx.coroutines.CancellationException
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.channels.ClosedReceiveChannelException
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.flow
import kotlinx.coroutines.isActive
import kotlinx.coroutines.launch
import net.sergeych.kiloparsec.AsyncVarint
import net.sergeych.kiloparsec.LocalInterface
import net.sergeych.mp_logger.*
import net.sergeych.tools.AtomicCounter

class SocketNetworkAddress(override val host: String, override val port: Int) : NetworkAddress {
    override fun toString(): String {
        return "$host:$port"
    }
}

actual fun NetworkAddress(host: String, port: Int): NetworkAddress = SocketNetworkAddress(host, port)

private val logCounter = AtomicCounter(0)

class ProtocolException(text: String, cause: Throwable? = null) : RuntimeException(text, cause)

const val MAX_TCP_BLOCK_SIZE = 16776216

actual fun acceptTcpDevice(port: Int): Flow<InetTransportDevice> {
    val selectorManager = SelectorManager(Dispatchers.IO)
    val serverSocket = aSocket(selectorManager).tcp().bind("127.0.0.1", port)
    val log = LogTag("TCPS${logCounter.incrementAndGet()}")
    return flow {
        while(true) {
            log.info { "Accepting incoming connections on $port" }
            serverSocket.accept().let { sock ->
                log.info { "Emitting transport device" }
                emit(inetTransportDevice(sock, "srv"))
            }
        }
    }
}

actual suspend fun connectTcpDevice(address: NetworkAddress): InetTransportDevice {
    val selectorManager = SelectorManager(Dispatchers.IO)
    val socket = aSocket(selectorManager).tcp().connect(address.host, address.port)
    println("Connected to ${address.host}:${address.port}")
    return inetTransportDevice(socket)
}


private fun inetTransportDevice(
    sock: Socket,
    suffix: String = "cli",
): InetTransportDevice {
    val networkAddress = sock.remoteAddress.toJavaAddress().let { NetworkAddress(it.hostname, it.port) }
    val inputBlocks = Channel<UByteArray>(4096)
    val outputBlocks = Channel<UByteArray>(4096)

    val log = LogTag("TCPT${logCounter.incrementAndGet()}:$suffix:$networkAddress")

    fun stop() {
        log.info { "stopping" }
        runCatching { inputBlocks.close()}
        runCatching { outputBlocks.close()}
        if( !sock.isClosed ) runCatching { sock.close()}
    }

    sock.launch {
        log.debug { "opening read channel" }
        val sockInput = runCatching { sock.openReadChannel() }.getOrElse {
            log.warning { "failed to open read channel $it" }
            sock.close()
            throw IllegalStateException("failed to open read channel")
        }
        while (isActive && sock.isActive) {
            try {
                val size = AsyncVarint.decodeUnsigned(sockInput).toInt()
                if (size > MAX_TCP_BLOCK_SIZE) // 16M is a max command block
                    throw ProtocolException("Illegal block size: $size should be < $MAX_TCP_BLOCK_SIZE")
                log.info { "read size: $size" }
                val data = ByteArray(size)
                log.info { "data ready" }
                sockInput.readFully(data, 0, size)
                inputBlocks.send(data.toUByteArray())
            } catch (e: ClosedReceiveChannelException) {
                stop()
                break
            } catch (_: CancellationException) {
            } catch (e: Exception) {
                log.exception { "unexpected exception in TCP socket read" to e }
                stop()
                break
            }
        }
    }
    sock.launch {
        val sockOutput = sock.openWriteChannel()
        while (isActive && sock.isActive) {
            try {
                val block = outputBlocks.receive()
                AsyncVarint.encodeUnsigned(block.size.toULong(), sockOutput)
                sockOutput.writeFully(block.toByteArray(), 0, block.size)
                log.info { "Client sock output: ${block.size}" }
                sockOutput.flush()
            } catch (_: CancellationException) {
                log.info { "Caught cancellation, closing transport" }
            } catch (_: LocalInterface.BreakConnectionException) {
                log.info { "requested connection break" }
                stop()
                break
            } catch (_: ClosedReceiveChannelException) {
                log.info { "receive block channel closed, closing the socket" }
                stop()
                break
            } catch (e: Exception) {
                log.exception { "unexpected exception. closing." to e }
                stop()
                break
            }
        }
    }
    val device = InetTransportDevice(inputBlocks, outputBlocks, networkAddress, {
        log.info { "Close has been called" }
        stop()
    })
    log.info { "Transport ready" }
    return device
}
